/* This is my Fast-Fractal-Algorithm ported from my own MC68000 assembly
 * routine done for the QL between April 1989 and July 1992, which is the
 * fastest zoomable algorithm ever seen on a QL - no doubt ;-)
 *
 * The algorithm uses a fixed-point format, which is formatted as follows:
 *
 *	   3	      2		 1
 *	# 1098765 43210987 654321098 76543210
 *	  sbbaaaa aaaaaaaa aaaaaaaaa aaaaaaaa
 *	  |||++++ ++++++++ +++++++++ ++++++++-> after decimal point
 *	  |++---- -------- --------- ---------> before decimal point
 *	  +------ -------- --------- ---------> sign bit (2's complement)
 *
 * The resolution is 2^-29 == 1.862645149 * 10^-9
 * The decimal range is -3.999999999 .. +3.999999999
 *
 * Algorithm:
 *   z = z^2 + c { z and c are complex numbers }
 *   z starts with 0 + 0i
 *   c are the coordinates x + yi
 *
 * void
 * asm_draw_mandel( itc, upx, y, r_max, i_val, d_real, d_imag)
 *   int  itc,		* iteration count
 *	  upx,		* column (right-most)
 *	  y,		* row (bottom-line)
 *	  r_max,	* real value
 *	  i_val,	* imag value
 *	  d_real,	* delta real
 *	  d_imag;	* delta imag
 *
 * stack layout:
 *   28+28 == 56(%esp) -> d_imag
 *   28+24 == 52(%esp) -> d_real
 *   28+20 == 48(%esp) -> i_val
 *   28+16 == 44(%esp) -> r_max
 *   28+12 == 40(%esp) -> y
 *   28+08 == 36(%esp) -> upx
 *   28+04 == 32(%esp) -> itc			* >  parameters
 *   28+00 == 28(%esp) -> return address	* == link
 *   28-04 == 24(%esp) -> %ebx			* <  caller-save registers
 *   28-08 == 20(%esp) -> %ebp
 *   28-12 == 16(%esp) -> %edi
 *   28-16 == 12(%esp) -> %esi
 *   28-20 ==  8(%esp) -> x			* <  local variable
 *   28-24 ==  4(%esp) -> par.y			* <  function parameters
 *   28-28 ==   (%esp) -> par.x or color
 *
 * register usage:
 *   %ebx -> real
 *   %ebp -> imag
 *   %ecx -> iter
 *   %edi -> real val
 *   %esi -> imag val
 *   %eax, %edx -> needed for multiplication #-(
 *
 * Note:  while reading the comments below keep in mind, that instructions
 *	  are scheduled, where possible, to omit penalty clock's on a 486er
 */

.text
#ifndef	__ELF__
	.align	4
.globl	_asm_draw_mandel
_asm_draw_mandel:
#else
	.align	16
.globl	asm_draw_mandel
	.type	asm_draw_mandel,@function
asm_draw_mandel:
#endif
	pushl	%ebx		/* preserve caller-save registers */
	pushl	%ebp
	pushl	%edi
	pushl	%esi
	subl	$12,%esp	/* 1 int and 2 parameters needed */
	movl	48(%esp),%esi	/* %esi= i_val */
	movl	52(%esp),%eax	/* r_max += d_real; */
	subl	56(%esp),%esi	/* i_val -= d_imag; */
	addl	%eax,44(%esp)
y_loop:
	addl	56(%esp),%esi	/* i_val += d_imag; */
	movl	36(%esp),%eax	/* x= upx */
	movl	44(%esp),%edi	/* %edi= r_val */
	movl	%eax,8(%esp)
x_loop:
	subl	52(%esp),%edi	/* r_val -= d_real; */
	movl	%esi,%ebp	/* %ebp= imag= i_val */
	movl	%edi,%ebx	/* %ebx= real= r_val */
pre_iter_begin:
	movl	%edi,%eax	/* pre-evaluate the sequence located
				 * between the two labels
				 * `iter_begin' and `iter_exit'
				 */
	movl	$1,%ecx		/* %ecx= iter= 1- itc
				 * as we start with z = z^2 + c
				 * with z = 0 + 0i and c = x +yi
				 * the first iteration will result in
				 * z = 0^2 +c = c = x + yi
				 */
	imull	%eax
	subl	32(%esp),%ecx
	sall	$3,%edx
	jo	iter_exit
	jc	iter_exit
	movl	%ebp,%eax
	addl	%ebx,%ebp
	movl	%edx,%ebx
	imull	%eax
	sall	$3,%edx
	jo	iter_exit
	jc	iter_exit
	addl	%ebx,%edx
	jo	iter_exit
iter_repeat:
	incl	%ecx		/* if ( ++iter == 0)  break; */
	jz	color_zero	/* assume it's inside, so "paint it black"
				 * ahem, any similarity with that famous
				 * Stones' song is purely random ;-)
				 */
	movl	%ebp,%eax	/* %eax= real+ imag
				 *
				 * %ebx= real^2- imag^2+ imag_val
				 *     = 2* real^2- ( real^2+ imag^2)
				 *	 + imag_val
				 */
	leal	(%edi,%ebx,2),%ebx
	movl	%esi,%ebp
	subl	%edx,%ebx
				/* %ebp= 2* real* imag+ imag_val
				 *     = imag_val- ( real^2+ imag^2)
				 *	 + ( real+ imag)^2
				 */
	subl	%edx,%ebp
	imull	%eax		/* ( real+ imag)^2 */
	leal	(%ebp,%edx,8),%ebp
iter_begin:
	movl	%ebx,%eax	/* %eax= real */
	imull	%eax
	sall	$3,%edx		/* if ( real^2 >= 4)  break;
				 * correctly this should be a
				 * `shldl $3,%eax,%edx', but
				 * linux-gas v2.5.2.6 doesn't recognize
				 * the immediate parameter as a valid one #-(
				 * the only solution would be to setup %ecx
				 * with the value 3, but that would mean this
				 * register isn't free as the iteration-counter
				 * any longer - not worth for the extra
				 * precision of 5.587935448 * 10^-9 IMHO
				 */
	jo	iter_exit
	jc	iter_exit	/* this 2nd j<cc> is necessary, because
				 * Intel chip designer's doesn't know the
				 * difference between an arithmetical and
				 * a logical left-shift %-6
				 * (hint: it's a matter of how to set the
				 *	  V condition bit)
				 */
	movl	%ebp,%eax	/* %eax= imag */
	addl	%ebx,%ebp	/* %ebp= real+ imag */
	movl	%edx,%ebx	/* %ebx= real^2 */
	imull	%eax
	sall	$3,%edx		/* if ( imag^2 >= 4)  break; */
	jo	iter_exit
	jc	iter_exit	/* see comments above ! */
	addl	%ebx,%edx	/* if ( real^2+ imag^2 >= 4)  break; */
	jno	iter_repeat
iter_exit:
	addl	32(%esp),%ecx	/* iter= itc- iter
				 * we'll use 129 colors, black for all values
				 * inside the Mandelbrot set and 128 "speed"
				 * measuring colors
				 */
	andl	$127,%ecx	/* color= ( iter % 128)+ 1; */
	incl	%ecx		/* skip over black */
color_zero:
	movl	40(%esp),%eax	/* %eax= y */
	movl	8(%esp),%ebp	/* %ebp= x */
	movl	%eax,4(%esp)	/* "push" parameter y */
	movl	%ecx,(%esp)	/* vga_setcolor( color); */
#ifndef	__ELF__
	call	_vga_setcolor
#else
	call	vga_setcolor
#endif
	movl	%ebp,(%esp)	/* vga_drawpixel( x, y); */
#ifndef	__ELF__
	call	_vga_drawpixel
#else
	call	vga_drawpixel
#endif
x_end:
	decl	8(%esp)		/* do..while ( --x >= 0); */
	jge	x_loop
y_end:
	decl	40(%esp)	/* do..while ( --y >= 0); */
	jge	y_loop
	addl	$12,%esp	/* remove 2 parameters and 1 int */
	popl	%esi		/* retrieve caller-save registers */
	popl	%edi
	popl	%ebp
	popl	%ebx
	ret
#ifdef	__ELF__
end:
	.size	asm_draw_mandel,end-asm_draw_mandel
#endif
